// Hey Emacs! This is -*- c++ -*- ! :-)
// domain.cc	-- Hans-Peter Stoerr Wed Jan 27 1999

#include <algo.h>
#include <functional>

#include "general.h"

#include "Domain.h"
#include "pddldomain.h"
#include "Term_func.h"
#include "misc/mystl.h"
#include "Substitution.h"
#include "Action.h"
#include "FluentIterator.h"

Option typeordered("typeordered","Use type variable ordering heuristic.",1);

Option movetypes("movetypes","adapt types of objects in problem to argument types of fluents",0);

// ================================================================
// misc helper functions

void    sortvectors(StringStringVectorMap& map)
{ // we sort the vectors and make sure every element is unique.
  for (StringStringVectorMap::iterator it=map.begin();
       map.end()!=it; ++it) {
    sort(it->second.begin(), it->second.end());
    StringVector::iterator uniquelast=unique(it->second.begin(), it->second.end());
    it->second.erase(uniquelast,it->second.end());
  }
}

Bdd Domain::StripsComplete(Bdd eff)
  // add identity if no change for all fluents.
{
  if (eff==mgr.False()) return eff;
  BddVector support=eff.Support(); // this still includes whens.
  BddVector realsupport;        // all possible effects
  BddVector res;
  res.push_back(eff);
  BddVector::const_iterator it(support.begin());
  for(; support.end()!=it; it++){
    Bdd fluent_primed=*it;
    string fname=fluent_primed.VarName();
    if (*(fname.end()-1)=='`') {
      realsupport.push_back(*it);
      fname.erase(fname.end()-1);
      Bdd fluent=mgr.Var(fname);
      res.push_back(eff.ForAll(fluent_primed).Implies( (fluent).Equiv(fluent_primed) ));
    }
  }
  Bdd result=mIdent.Exists(realsupport)*MultiwayAnd(res);
  return result;
}

Bdd MaxNOf(int nmax,BddVector v) 
{
  int n=0,i,siz=v.size();
  BddVector     vn,vn1;
  // vn[i+1]=at most n of v[0..i]
  if (siz==0) return mgr.One();
  if (nmax>=siz) return mgr.One();
  for (i=0;siz>=i;++i) vn1.push_back(mgr.Zero());
  for (n=0;nmax>=n;++n) {
    vn=vn1;
    vn1[0]=mgr.One(); // there are always at most n>0 of one variable
    // cout << "MaxNOf(" << n << "," << 0 << ")= " << vn1[0] << "\n";
    for (i=1;siz-nmax+n>=i;i++) {
      vn1[i]=v[i-1]*vn[i-1]+(-v[i-1])*vn1[i-1];
      // cout << "MaxNOf(" << n << "," << i << ")= " << vn1[i] << "\n";
    }
  }
//   Bdd res=vn1.back();
//   while (mgr.Zero()!=res) {
//     Bdd cub=res.Satisfy();
//     printit(cub);
//     res=res*-cub;
//   }
  // printit(nmax);
  // printit(v);
  // printit(vn1.back());
  return vn1.back();
}

// ================================================================
// Domain Implementation:
// Lifecycle

Domain::Domain(const PddlDatabase& database, string domain, string problem)
  : prob(*database.Problem(domain,problem)),
    dom(*database.Domain(database.Problem(problem)->domainname))
{
  *gpLog << "Problem: " << prob.name << " from domain "
         << dom.filename << ":" << dom.name << ".\n";
}

Domain::~Domain()
{
  NOOP;
}

void    Domain::Printit(ostream& s) const
{
  s << "Domain:"
    << "\n mTypeConstantsMap:\t" << mTypeConstantsMap
    << "\n mEffectFnameAname:\t" << mEffectFnameAname
    << "\n mPrecondFnameAname:\t" << mPrecondFnameAname
    << "\n mConstantFnames:\t" << mConstantFnames
    << "\n mConstantFluents\t" << mConstantFluents
    << "\n mConstantValues:\t" << mConstantValues
    << "\n mFluents:\t" << mFluents
    << "\n mIdent:\t" << mIdent
    << "\n mTimeless:\t" << mTimeless
    << "\n mAxioms:\t" << mAxioms
    << "\n mConstraints:\t" << mConstraints
    << "\n mFilter:\t" << mFilter
    << "\n mInit:\t" << mInit
    << "\n mGoal:\t" << mGoal
    << "\n.\n";
}

// ================================================================
// now some actual processing is done.

void    Domain::Init()
  // Wrapper which calls preprocessing functions.
{
  CollectConstants();
  MakeFnameAnameMaps();
  MakeConstantFnames();
  MakeBddVarOrder();
  MakeTimeless();
  MakeAxioms();
  MakeInitialState();
  MakeGoal();
  MakeConstraints();
  MakeFilter();
  MakeIdentityBdd();
}

// ________________________________________________________________
// collection of constants into mTypeConstantsMap.

class DomainCollectConstantsFunc : public RecursiveVoidTermFunction
{
public:
  DomainCollectConstantsFunc(StringStringMap& theconsts,
                             const StringAtomtypeMap& theatomtypes,
                             const StringStringMap& thedirectsupertypes)
    : consts(theconsts), atomtypes(theatomtypes), directsupertypes(thedirectsupertypes) {}
protected:
  StringStringMap&      consts;
  const StringAtomtypeMap&    atomtypes;
  const StringStringMap&        directsupertypes;
  
  virtual void  AtomFunc(const TermAtom&);
  void  PossiblyInsertConstant(string,string);
};

void    DomainCollectConstantsFunc::AtomFunc(const TermAtom& t)
{
  StringAtomtypeMap::const_iterator it=atomtypes.find(t.name);
  if (atomtypes.end()==it) {
    *gpLog << "PDDL bug: Atom " << t.name << " not found.\n" << flush;
    abort();
  }
  const Atomtype& type=it->second;
  for_each2(t.args.begin(), t.args.end(), type.begin(), type.end(),
              bindthis(&DomainCollectConstantsFunc::PossiblyInsertConstant,this));
}

void    DomainCollectConstantsFunc::PossiblyInsertConstant
(string constant, string type)
{ // check if a constant of name constant is already defined. If not insert if with type.
  StringStringMap::iterator constpos=consts.find(constant);
  if (consts.end()==constpos) { // constant not defined.
    consts[constant]=type; return; // define it and done.
  }
  if (type!=constpos->second && movetypes) { // we check if the type the constant is used gives us 
    // a new constraint on the type of the used constant.
    // First see if type is a supertype of the type the object has now.
    StringStringMap::const_iterator it=directsupertypes.find(constpos->second),
      end=directsupertypes.end(); // now go up hierarchy till top or found.
    while ( end!=it && it->second!=type ) it=directsupertypes.find(it->second);
    if ( end==it ) { // we haven't found type, so we assume it is more specific
      // check now if type really is a subtype of type it has now.
      it=directsupertypes.find(type);
      while ( end!=it && it->second!=constpos->second ) it=directsupertypes.find(it->second);
      if ( end==it ) {
        *gpLog << "PDDL bug: type " << type << " inferred for " << constant
               << " but is not a subtype of " << constpos->second << " .\n" << flush;
        abort();
      }
      constpos->second=type;
    }
    // if we have found type we are already happy.
  }
}

static StringVector
sGenerateEitherType(string eithertype,
                    const StringStringVectorMap& mTypeConstantsMap)
{
  StringVector res,empty;
  string::size_type pre,post;
  pre=eithertype.find("(");
  post=eithertype.find_first_of(",)",pre+1);
  do {
    StringVector res1=res;
    string type(eithertype,pre+1,post-pre-1);
    res=empty;
    CHECK(mTypeConstantsMap.end()!=mTypeConstantsMap.find(type),
          "Type "+type+" not found.");
    const StringVector& typeobjs=mTypeConstantsMap.find(type)->second;
    set_union(res1.begin(), res1.end(),
              typeobjs.begin(),typeobjs.end(),
              back_inserter(res));
    pre=post;
    post=eithertype.find_first_of(",)",pre+1,2);
  } while (')'!=eithertype[pre]);
  return res;
}

static StringSet SplitEitherType(string eithertype)
{
  StringSet    res;
  string::size_type pre,post;
  pre=eithertype.find("(");
  post=eithertype.find_first_of(",)",pre+1,2);
  do {
    string type(eithertype,pre+1,post-pre-1);
    res.insert(type);
    pre=post;
    post=eithertype.find_first_of(",)",pre+1,2);
  } while (')'!=eithertype[pre]);
  return res;
}

void    Domain::CollectConstants()
{
  StringStringMap       consts=dom.constants;
  // now we collect constants from the problem.
  copy(prob.constants.begin(), prob.constants.end(),
       inserter(consts,consts.end()));

  // perhaps we have to infer new constants from the initial situation.
  DomainCollectConstantsFunc  collectConst(consts,dom.atomtypes,dom.directsupertypes);
  for_each(prob.init.begin(), prob.init.end(), collectConst);

  // add constants to types and supertypes in mTypeConstantsMap
  for(StringStringMap::iterator cit=consts.begin();
      consts.end()!=cit; ++cit){
    string type=cit->second;
    if (type.find("either(")!=0) 
      do {
        mTypeConstantsMap[type].push_back(cit->first);
        if (dom.directsupertypes.end()!=dom.directsupertypes.find(type))
          type=dom.directsupertypes.find(type)->second;
        else type="";
      } while(""!=type);
    else { // if the type starts with either( ...
      StringSet types(SplitEitherType(type));
      StringSet::const_iterator itt=types.begin(),endt=types.end();
      for (;endt!=itt;++itt) {
        type=*itt;
        do {
          mTypeConstantsMap[type].push_back(cit->first);
          if (dom.directsupertypes.end()!=dom.directsupertypes.find(type))
            type=dom.directsupertypes.find(type)->second;
          else type="";
        } while(""!=type);
      }
    }
  }

  // last we have to generate either(...) types occuring in the atomtypes.
  string either("either(");
  // YUK! Warum funktioniert StringAtomtypeMap::iterator hier nicht?
  typeof(dom.atomtypes.begin()) atmit;
  Atomtype::const_iterator atit;
  for(atmit=dom.atomtypes.begin(); dom.atomtypes.end()!=atmit; ++atmit)
    for(atit=atmit->second.begin(); atmit->second.end()!=atit; ++atit) {
      if (0==atit->compare(0,7,either) &&
          mTypeConstantsMap.end()==mTypeConstantsMap.find(*atit))
        mTypeConstantsMap[*atit]=sGenerateEitherType(*atit,mTypeConstantsMap);
    }
  sortvectors(mTypeConstantsMap);
}

// ________________________________________________________________
// collect preconditions and postconditions of actions

class DomainCollectPrecond: public RecursiveVoidTermFunction
{
public:
  DomainCollectPrecond(string theaname,StringStringVectorMap& thepremap)
  : aname(theaname), mPrecondFnameAname(thepremap) {};

protected:
  string aname;                 // action name we work on 
  StringStringVectorMap& mPrecondFnameAname; // where to collect the precond.
  virtual void  AtomFunc(const TermAtom&);
  virtual void  SpecialFunc(const TermUnary&) {};
};

void DomainCollectPrecond::AtomFunc(const TermAtom& t)
{
  StringVector& fvec=mPrecondFnameAname[t.name];
  if ( 0==fvec.size() || fvec.back()!=aname )
    fvec.push_back(aname);
}

class DomainCollectEffects: public RecursiveVoidTermFunction
{
public:
  DomainCollectEffects(string theaname, StringStringVectorMap& theeffmap,
                       DomainCollectPrecond& thepcollector)
  : aname(theaname), mEffectFnameAname(theeffmap),
    pcollector(thepcollector) {};

protected:
  string aname;
  StringStringVectorMap& mEffectFnameAname;
  DomainCollectPrecond&  pcollector;

  virtual void  WhenFunc(const TermBinary&);
  virtual void  AtomFunc(const TermAtom&);
};

void DomainCollectEffects::AtomFunc(const TermAtom& t)
{
  StringVector& fvec=mEffectFnameAname[t.name];
  if ( 0==fvec.size() || fvec.back()!=aname )
    fvec.push_back(aname);
}

void DomainCollectEffects::WhenFunc(const TermBinary& t)
{
  Call(t.arg2);
  pcollector(t.arg1);
}

void  Domain::MakeFnameAnameMaps()
{
  StringPddlActionMap::const_iterator action;
  for(action=dom.actions.begin(); dom.actions.end()!=action; ++action) {
    DomainCollectPrecond pcollector(action->second.name,mPrecondFnameAname); // Huh? What did I do this for?
    DomainCollectEffects ecollector(action->second.name,mEffectFnameAname,pcollector);
    pcollector(action->second.precondition);
    StringStringVectorMap backup=mPrecondFnameAname; // safe this value; we don't want whens in this.
    ecollector(action->second.effect);
    mPrecondFnameAname=backup;
  }
  sortvectors(mPrecondFnameAname);
  sortvectors(mEffectFnameAname);
}

// ________________________________________________________________
// collect constant fluents

void Domain::MakeConstantFnames()
{
  StringAtomtypeMap::const_iterator it;
  for ( it=dom.atomtypes.begin(); dom.atomtypes.end()!=it; ++it) 
    if( mEffectFnameAname.end()==mEffectFnameAname.find(it->first) ) 
      mConstantFnames.insert(it->first);
}

bool Domain::IsConstantFluent(const string fluent)
{
  return IsConstantFname(fluent.substr(0,fluent.find("(")));
}

// ================================================================
// Going to Bdds

class TermToBddFunc : public TermBinaryFunction<Substitution,Bdd>
{
public:
  TermToBddFunc(const Domain& dom) : mDom(dom){};
  virtual       ~TermToBddFunc() {};
protected:
  const Domain&       mDom;
  
  virtual Bdd   NullFunc(Substitution s);
  virtual Bdd   NotFunc(const TermUnary&, Substitution s);
  virtual Bdd   AndFunc(const TermAc1Binary&, Substitution s);
  virtual Bdd   OrFunc(const TermAc1Binary&, Substitution s);
  virtual Bdd   ImplyFunc(const TermBinary&, Substitution s);
  virtual Bdd   ForallFunc(const TermQuantor&, Substitution s);
  virtual Bdd   ExistsFunc(const TermQuantor&, Substitution s);
  virtual Bdd   EqualFunc(const TermEqual&, Substitution s);

  Bdd   quantorIterate(StringStringMap::const_iterator begin,
                       StringStringMap::const_iterator end,
                       const Term t, const Substitution s, Bdd init,
                       Bdd (Bdd::* op)(const Bdd) const,
                       const Bdd opfixpoint);
};

Bdd     TermToBddFunc::NotFunc(const TermUnary& t, Substitution s)
{
  return -Call(t.arg,s);
}

Bdd     TermToBddFunc::AndFunc(const TermAc1Binary& t, Substitution s)
{
  Bdd res=mgr.One();
  TermSet::iterator it;
  for(it=t.args.begin(); t.args.end()!=it; ++it) {
    res*=Call(*it,s);
    if (mgr.Zero()==res) return res;
  }
  return res;
}
           
Bdd     TermToBddFunc::OrFunc(const TermAc1Binary& t, Substitution s)
{
  Bdd res=mgr.Zero();
  TermSet::iterator it;
  for(it=t.args.begin(); t.args.end()!=it; ++it) {
    res+=Call(*it,s);
    if (mgr.One()==res) return res;
  }
  return res;

}

Bdd     TermToBddFunc::ImplyFunc(const TermBinary& t, Substitution s)
{
  Bdd argOne=Call(t.arg1,s);
  if (mgr.Zero()==argOne) return mgr.One();
  return -argOne+Call(t.arg2,s);
}

Bdd     TermToBddFunc::ForallFunc(const TermQuantor& t, Substitution s)
{
  return quantorIterate(t.vars.begin(),t.vars.end(),t.arg,s,
                        mgr.One(), &Bdd::operator*, mgr.Zero());
}

Bdd     TermToBddFunc::ExistsFunc(const TermQuantor& t, Substitution s)
{
  return quantorIterate(t.vars.begin(),t.vars.end(),t.arg,s,
                        mgr.Zero(), &Bdd::operator+, mgr.One());
}

Bdd     TermToBddFunc::
quantorIterate(StringStringMap::const_iterator begin,
               StringStringMap::const_iterator end,
               const Term t, const Substitution s,
               Bdd init, Bdd (Bdd::* op)(const Bdd) const,
               const Bdd fixpoint)
{
  if (begin==end) return (init.*op)(Call(t,s));
  Substitution sn(s);
  string var=begin->first;
  if (mDom.mTypeConstantsMap.find(begin->second)==mDom.mTypeConstantsMap.end()) // set of objects empty
    return -fixpoint;
  const StringVector& objects=mDom.mTypeConstantsMap.find(begin->second)->second;
  StringVector::const_iterator it;
  begin++;
  for(it=objects.begin(); objects.end()!=it; ++it){
    sn[var]=*it;
    init=quantorIterate(begin,end,t,sn,init,op,fixpoint);
    if (fixpoint==init) return init;
  }
  return init;
}

Bdd     TermToBddFunc::EqualFunc(const TermEqual& t, Substitution s)
{
  if (s(t.arg1)==s(t.arg2)) return mgr.One();
  else return mgr.Zero();
}

Bdd     TermToBddFunc::NullFunc(Substitution) { IMPOSSIBLE; return mgr.Zero(); }

// ________________________________________________________________

class TermToBddPrecond : public TermToBddFunc
{
public:
  TermToBddPrecond(const Domain& dom) : TermToBddFunc(dom) {};
  virtual       ~TermToBddPrecond() {};
protected:
  virtual Bdd   WhenFunc(const TermBinary&, Substitution s);
  virtual Bdd   AtomFunc(const TermAtom&, Substitution s);
  virtual Bdd   SpecialFunc(const TermUnary&, Substitution s);
  virtual Bdd   NullFunc(Substitution s);
private:
  TermToBddPrecond(const TermToBddPrecond&); // undefined
  TermToBddPrecond& operator=(const TermToBddPrecond&); // undefined
};

Bdd     TermToBddPrecond::WhenFunc(const TermBinary&, Substitution)
{
  IMPOSSIBLE;                   // when shouldn't occur in preconditions
  return mgr.Zero();
}

Bdd     TermToBddPrecond::AtomFunc(const TermAtom& t, Substitution s)
{
  string aname=t.name+"(";
  StringList::const_iterator it=t.args.begin();
  if (t.args.end()!=it) aname=aname+s(*(it++));
  while (t.args.end()!=it) aname=aname+","+s(*(it++));
  aname=aname+")";
  if ( mDom.mConstantFnames.end()!=mDom.mConstantFnames.find(t.name) ||
       mDom.mConstantFluents.end()!=mDom.mConstantFluents.find(aname) )
    // fluent is a constant
    return mDom.mConstantValues(aname);
  return mgr.Var(aname);
}

Bdd     TermToBddPrecond::SpecialFunc(const TermUnary& t, Substitution s)
{
  Bdd   arg=Call(t.arg,s);
  if ("IsGoal"==t.name) {
    if ( mgr.One()==mDom.mGoal.Implies(arg) ) // if argument is always true in goal
      return mgr.One();
    return mgr.Zero();
  }
  if ("IsInitial"==t.name) {
    // printit(mDom.mInit);
    // printit(mDom.mInit.Implies(arg));
    if ( mgr.One()==mDom.mInit.Implies(arg) ) // if argument is always true in initial state
      return mgr.One();
    return mgr.Zero();
  }
  if (0==t.name.find("Limit")) {
    return MaxNOf(atoi(t.name.substr(5).c_str()),arg.Support());
  }
  UNIMPLEMENTED;
}

Bdd     TermToBddPrecond::NullFunc(Substitution)
{
  return mgr.One();             // an empty precondition is true.
}

// ________________________________________________________________

class TermToBddEffect : public TermToBddFunc
{
public:
  TermToBddEffect(const Domain& dom, TermToBddPrecond& prefunc)
    : TermToBddFunc(dom), mPreFunc(prefunc) {};
  virtual       ~TermToBddEffect() {};
protected:
  TermToBddPrecond& mPreFunc;
  
  virtual Bdd   WhenFunc(const TermBinary&, Substitution s);
  virtual Bdd   AtomFunc(const TermAtom&, Substitution s);
  virtual Bdd   SpecialFunc(const TermUnary&, Substitution s);
  virtual Bdd   NullFunc(Substitution s);
private:
  TermToBddEffect(const TermToBddEffect&); // undefined
  TermToBddEffect& operator=(const TermToBddEffect&); // undefined
};

Bdd     TermToBddEffect::WhenFunc(const TermBinary& t, Substitution s)
{
  return -mPreFunc(t.arg1,s)+Call(t.arg2,s);
}

Bdd     TermToBddEffect::AtomFunc(const TermAtom& t, Substitution s)
{
  string aname=t.name+"(";
  StringList::const_iterator it=t.args.begin();
  if (t.args.end()!=it) aname=aname+s(*(it++));
  while (t.args.end()!=it) aname=aname+","+s(*(it++));
  aname=aname+")";
  if ( mDom.mConstantFnames.end()!=mDom.mConstantFnames.find(t.name) ||
       mDom.mConstantFluents.end()!=mDom.mConstantFluents.find(aname) )
    // fluent is a constant
    return mDom.mConstantValues(aname);
  return mgr.Var(aname+"`");
}

Bdd     TermToBddEffect::SpecialFunc(const TermUnary&, Substitution)
{
  IMPOSSIBLE;                   // Specials shouldn't occur in effects.
  return mgr.Zero();
}

Bdd     TermToBddEffect::NullFunc(Substitution)
{
  return mgr.Zero();            // :TODO: What is an empty effect? Maybe false.
}

// ________________________________________________________________

Bdd Domain::TermToBdd(Term t,Substitution s) const
{
  TermToBddPrecond makeBdd(*this);
  return makeBdd(t,s);
}

Bdd Domain::EffectTermToBdd(Term t, Substitution s) const
{
  TermToBddPrecond makePreBdd(*this);
  TermToBddEffect makeEffBdd(*this,makePreBdd);
  return makeEffBdd(t,s);
}

void Domain::MakeGoal()
{
  mGoal=TermToBdd(prob.goal);
}

void Domain::MakeTimeless()
{
  mTimeless=mgr.One();
  if (0<dom.timeless.size())
    mTimeless=TermToBdd(MakeTerm(Term::And,dom.timeless));
}

void Domain::MakeAxioms()
{
  mAxioms=mgr.One();
  CHECK(0==dom.axioms.size(), "Axioms not supported yet.");
}

void Domain::MakeConstraints()
{
  mConstraints=mgr.One();
  if (0<dom.constraints.size())
    mConstraints=TermToBdd(MakeTerm(Term::And,dom.constraints));
}

void Domain::MakeFilter()
{
  mFilter=mConstraints*mAxioms;
  Bdd a("a"),b("b");
  Bdd   newconstants=mFilter.SuperCube();
  BddVector     theconstants(newconstants.Support());
  if (0==theconstants.size()) return; // return if no new constants introduced.
  BddAssoc      noconstants(theconstants);
  mFilter=mFilter.Exists(noconstants);
  CHECK(mgr.Zero()!=mFilter*mInit , "Constraints conflict with initial state.");
  mInit=(mFilter*mInit).Exists(noconstants);
  mGoal=(mFilter*mGoal).Exists(noconstants);
  BddVector::const_iterator it(theconstants.begin());
  const BddVector::const_iterator end(theconstants.end());
  for (;end!=it;++it) {
    mConstantFluents.insert(it->VarName());
    if (mgr.Zero()!=newconstants*(*it)) mConstantValues.insert(it->VarName());
  }
  // now remove new constants from mFluents.
  StringVector     mflunew;        // new value for mFluents
  StringVector::const_iterator itflu=mFluents.begin(),endflu=mFluents.end();
  for (;endflu!=itflu; ++itflu) 
    if ( !newconstants.DependsOn(mgr.Var(*itflu)) ) // only add fluent if it is not constant.
      mflunew.push_back(*itflu);
  mFluents=mflunew;
}

void Domain::MakeInitialState()
  // initializes mInit. call after MakeTimeless, MakeAxioms, MakeConstraints and MakeFilter!
{
  // initialize mConstantValues
  TermSet::const_iterator it; // insert values given in initial sit.
  for (it=prob.init.begin(); prob.init.end()!=it; ++it){
    if (Term::Atom==it->Type() &&
        mConstantFnames.end()!=mConstantFnames.find((*it)->name))
      mConstantValues.insert(it->Describe());
  }
  // now insert values for pseudoatoms as well
  StringStringVectorMap::const_iterator ittype=mTypeConstantsMap.begin(),
    endtype=mTypeConstantsMap.end();
  for (;endtype!=ittype; ++ittype) {
    StringVector::const_iterator it=ittype->second.begin(),
      end=ittype->second.end();
    for (;end!=it;++it)
      mConstantValues.insert(ittype->first+"("+*it+")");
  }

  // initialize mInit
  Bdd   res;                    // we put mInitult here first
  res=mgr.One();
  for(TermSet::iterator tit=prob.init.begin(); prob.init.end()!=tit; ++tit)
    res*=TermToBdd(*tit);
  res*=mTimeless;
  // add negation by default to mInit
  StringVector::const_iterator fit=mFluents.begin(),end=mFluents.end();
  for(; end!=fit; fit++ ){
    Bdd var(mgr.Var(*fit));
    if ( ! res.DependsOn(var) ) res*=-var;
  }
  mInit=res;
}

// ________________________________________________________________

struct sortbytypesizecmpfunc {
  const map<string,int> *ptype_size;
  const Atomtype *pftype;
  sortbytypesizecmpfunc(const map<string,int> *theptype_size, const Atomtype *thepftype)
    : ptype_size(theptype_size), pftype(thepftype) {}
  bool    operator()(int x,int y)
  { return ptype_size->find((*pftype)[x])->second > ptype_size->find((*pftype)[y])->second; }
};

static bool  sortbyfirstcmpfunc
(const pair<string,string> &x, const pair<string,string> &y)
{ return x.first<y.first; }

void Domain::MakeBddVarOrder() { // initialize variable order for Bdds
  if (!typeordered) {
    FluentIterator fbegin(*this),fend;
    copy(fbegin,fend,back_inserter(mFluents));
    StringVector::const_iterator sit=mFluents.end();
    while (sit!=mFluents.begin()) {
      sit--;
      mgr.Var(*sit);              // just generate variable
      mgr.Var(*sit+"`");              // just generate variable
      cout << *sit << ": " << mgr.Var(*sit).IfIndex() << "\n";
    }
    return;
  }
  // type ordering. Fluents are ordered by the object of their biggest objecttype,
  // then their second biggest type, etc., and then their name.
  // First we count the objects in each type.
  map<string,int> type_size;
  StringStringVectorMap::const_iterator typeiter=mTypeConstantsMap.begin();
  for(; mTypeConstantsMap.end()!=typeiter; ++typeiter)
    type_size[typeiter->first]=typeiter->second.size();
  FluentIterator fit(*this),fend;
  vector<int> argorder;
  vector<pair<string,string> > orderedfluents;
  for (; ! (fend==fit); ++fit) {
    if (fit.IsFirstOfType()) {
      // For each fluenttype we sort its argument numbers by type_size into args
      argorder=vector<int>(fit.Ftype().size());
      iota(argorder.begin(),argorder.end(),0);
      sort(argorder.begin(),argorder.end(),
           sortbytypesizecmpfunc(&type_size,&fit.Ftype()));
    }
    string ordername="(";
    FirsttimeNexttime<string> sep("",",");
    vector<int>::const_iterator ait;
    for (ait=argorder.begin(); argorder.end()!=ait; ++ait) {
      if (2==typeordered)
        ordername+=sep()+fit.Fargs()[*ait];
      else ordername+=sep()+fit.Ftype()[*ait]+":"+fit.Fargs()[*ait];
    }
    ordername+=")"+fit.Fname();
    orderedfluents.push_back(make_pair(ordername,*fit));
  }
  sort(orderedfluents.begin(), orderedfluents.end(), sortbyfirstcmpfunc);
  // now we initialize the variable order.
  vector<pair<string,string> >::const_iterator sit=orderedfluents.end();
  while (sit!=orderedfluents.begin()) {
    sit--;
    mgr.Var(sit->second);              // just generate variable
  }
  mFluents=StringVector();
  for(sit=orderedfluents.begin(); orderedfluents.end()!=sit; ++sit)
    mFluents.push_back(sit->second);
  return;
}

void Domain::MakeIdentityBdd()
  // does BDD describing identity of state and next state. Call after MakeFilter.
{
  mIdent=mgr.One();
  StringVector::const_iterator fit=mFluents.begin(),fend=mFluents.end();
  for (; fend!=fit; ++fit){
    string fname(*fit);
    mIdent*= mgr.Var(fname).Equiv(mgr.Var(fname+"`"));
  }
}

static bool sPrintBug=TRUE;
static bool sPrintContradictEffect=TRUE;

extern Option reduceactionset,actionsort;

pair<Bdd,BddVector>   Domain::ActionBdd(const Action& a)
{ // generates BDD for an action.
  // first component of pair is Bdd for Action;
  // second component is set of fluents that are relevant for precondition and effects
  Bdd precond,effect;
  pair<Bdd,BddVector> res;
  Term precondTerm,effectTerm;
  precondTerm=dom.actions.find(a.mName)->second.precondition;
  effectTerm=dom.actions.find(a.mName)->second.effect;
  const StringStringMap& vars=dom.actions.find(a.mName)->second.vars;
  if (0!=vars.size()) {
    effectTerm=MakeTerm(Term::Forall,vars,
                        MakeTerm(Term::When,precondTerm,effectTerm));
    precondTerm=MakeTerm(Term::Exists,vars,precondTerm);
  }
  precond=TermToBdd(precondTerm,a.mSubst);
  if (mgr.Zero()==precond) {
    res.first=precond; return res;
  }
  effect=EffectTermToBdd(effectTerm,a.mSubst);
  if (mgr.Zero()==effect && sPrintBug) {
    sPrintBug=FALSE; // print this just once.
    *gpLog << "PDDL bug: Effect of " << a << "is false.\n" << flush;
  }
  res.first=precond*StripsComplete(effect);
  if (sPrintContradictEffect && mgr.One()!=res.first.Exists(mgr.Primed())+ -precond) {
    sPrintContradictEffect=FALSE; // print this just once.
    *gpLog << "Warning: Action " << a.Description() << " has sometimes contradictory effects.\n";
  }
  Bdd reseff=precond*effect;
  res.second=reseff.Support();
  if (reduceactionset && !actionsort) {
    BddVector newsupp;
    BddVector::const_iterator it=res.second.begin(),end=res.second.end();
    for (; end!=it; ++it) {
      if (mgr.Zero()!=reseff*(*it)) newsupp.push_back(*it);
      if (mgr.Zero()!=reseff*(-*it)) newsupp.push_back(-*it);
    }
    res.second=newsupp;
  }
  return res;
}

// ================================================================

void Domain::TestFunc()
  // run some tests
{
//   ActionIterator it(*this),end;
//   copy(it,end,ostream_iterator<Action>(cout,"\n"));

//   Bdd effect,precond;
//   ActionIterator it(*this),end;
//   while( end!=it &&
//          mgr.Zero()== mInit*
//          TermToBdd(dom.actions.find(it->mName)->second.precondition,it->mSubst) )
//     it++;
//   it++;
//   while( end!=it &&
//          mgr.Zero()== mInit*
//          TermToBdd(dom.actions.find(it->mName)->second.precondition,it->mSubst) )
//     it++;
//   precond=TermToBdd(dom.actions.find(it->mName)->second.precondition,it->mSubst);
//   effect=EffectTermToBdd(dom.actions.find(it->mName)->second.effect,it->mSubst);
//   printit(*it);
//   printit(precond);
//   printit(effect);
//   BddAssoc state_assoc; state_assoc.SetToPrimedSet(0);
//   Bdd ac=(precond.Implies(StripsComplete(effect)));
//   Bdd step=mInit*ac;
//   printit(step);
//   printit(ac.Equiv(ActionBdd(*it)));
//   printit(step.Exists(state_assoc));
}
