/* domain.h -*- c++ -*- Hans-Peter Stoerr Wed Jan 27 1999 */

#ifndef _DOMAIN_H
#define _DOMAIN_H

#include <string>
#include <vector>
#include <map>
#include "Printable.h"
#include "objcal.h"
#include "Term.h"
#include "Interpretation.h"
#include "Substitution.h"

typedef set<string,less<string> > StringSet;
typedef vector<string> StringVector;
typedef map<string,StringVector,less<string> > StringStringVectorMap;

class PddlDatabase;
class PddlDomain;
class PddlProblem;
class Action;
class ActionIterator;

class Domain : public Printable
{
  friend class ActionIterator;
public:
  const PddlProblem&    prob;
  const PddlDomain&     dom;

  StringStringVectorMap mTypeConstantsMap;
  // tells Actions in which an Fluentname appears as effect:
  StringStringVectorMap mEffectFnameAname; // (sorted.)
  // tells Actions in which an Fluentname appears as precond.:
  StringStringVectorMap mPrecondFnameAname; // (sorted.)
  // set of names of constant fluents (i.e. they don't appear in effects)
  StringSet             mConstantFnames; // functors for which all fluents are constant.
  StringSet             mConstantFluents; // names of fluents which are constant (add. to mConstantFnames)
  Interpretation        mConstantValues; // Values of the constant fluents
  StringVector          mFluents; // all fluents in appropriate order.
  Bdd                   mIdent; // equivalence of state and next state
  Bdd                   mTimeless; // timeless fluents. Are just integrated into mInit -> never mind.
  Bdd                   mInit;  // Initial State
  Bdd                   mGoal;  // Goal State
  Bdd                   mFilter; // mAxioms and mConstraints put together
  
  Domain(const PddlDatabase& database, string domain, string problem);
  virtual ~Domain();

  void  Init();           // initialization of the domain.
  // returns Action execution BDD and set of fluents it changes or depends on:
  pair<Bdd,BddVector>   ActionBdd(const Action&);

  bool  IsConstantFname(const string fname)
  { return mConstantFnames.end()!=mConstantFnames.find(fname); }
  bool  IsConstantFluent(const string fluent);

  void  TestFunc();
protected:
  Bdd                   mAxioms; // contains all axioms.
  Bdd                   mConstraints; // contains all constraints

  void  CollectConstants();     // into mTypeConstantsMap
  void  MakeFnameAnameMaps();   // generate mPrecondFnameAname and mEffectFnameAname
  void  MakeConstantFnames();
  void  MakeBddVarOrder();      // initialize variable order for Bdds
  void  MakeIdentityBdd();      // initialize mIdent
  void  MakeTimeless();         // init mTimeless
  void  MakeAxioms();           // init mAxioms
  void  MakeConstraints();      // init mAxioms
  void  MakeFilter();           // init mFilter
  void  MakeInitialState();     // initialize mInit
  void  MakeGoal();             // init mGoal

  Bdd   StripsComplete(Bdd effect); // completion according to STRIPS-semantics
  
  // transform Bdd into Term.
  Bdd   TermToBdd(Term,Substitution s=Substitution()) const;
  Bdd   EffectTermToBdd(Term,Substitution) const; // transform action effect
  
  virtual void  Printit(ostream&) const;

private:
  Domain(const Domain&); // unimplemented, can't be done.
  Domain& operator=(const Domain&); // unimplemented, can't be done.
};

#endif
