// RubberbandOrder.cc -*- c++ -*- Hans-Peter Stoerr Fri Feb 11 2000
// generates a variable order by the "rubberband" principle.
// takes a set of edges (string,string),
// returns vector<string> (the order).

#include <assert.h>
#include <time.h>
#include <stdio.h>
#include <iostream.h>
#include "Printable.h"
#include "RubberbandOrder.h"
#include <algo.h>
#include "misc/mystl.h"

#ifndef linux
extern "C" 
{
  long random();
  int srandom( unsigned seed);
}
#endif
 
void    RubberbandOrder::InsertEdge(pair<string,string> inPair)
{
  assert(!mCalculated);
  int n1,n2;                    // the two node numbers
  n1=mNameToNumber[inPair.first]; n2=mNameToNumber[inPair.second];
  if(0==n1) {
    n1=++mCurNum; mNumNodes++;
    mNameToNumber[inPair.first]=n1; mNumberToName[n1]=inPair.first;
    mEdge.push_back(vector<int>()); // so mEdge contains as many el. as there are nodes.
  }
  if(0==n2) {
    n2=++mCurNum; mNumNodes++;
    mNameToNumber[inPair.second]=n2; mNumberToName[n2]=inPair.second;
    mEdge.push_back(vector<int>());// so mEdge contains as many el. as there are nodes.
  }
  if(mWeight.end()==mWeight.find(make_pair(n1,n2))) { // new edge
    mEdge[n1].push_back(n2); mEdge[n2].push_back(n1); mNumEdges++;
    mWeight[make_pair(n1,n2)]=1; mWeight[make_pair(n2,n1)]=1;
  }
  else { mWeight[make_pair(n1,n2)]++; mWeight[make_pair(n2,n1)]++; }
}

void    RubberbandOrder::MakeDistances(int n)
  // generate the nodes of each distance
{
  vector<int> stage;            // vectors of this stage.
  mDist.erase(mDist.begin(),mDist.end()); // reset computation vars.
  mConnected=vector<bool>(mNumNodes+1);
  stage.push_back(n); mConnected[n]=TRUE;
  mDist.push_back(stage);
  // now we leaf through the stages.
  while (0<mDist.back().size()) {
    stage=vector<int>();
    vector<int>::iterator curNode; // for every node of last stage...
    for (curNode=mDist.back().begin(); mDist.back().end()!=curNode; curNode++) {
      vector<int>::iterator connNode; // n. *curNode connects to
      for (connNode=mEdge[*curNode].begin();
           mEdge[*curNode].end()!=connNode; connNode++) {
          if (!mConnected[*connNode]) {
            mConnected[*connNode]=TRUE; // remember we had this node
            stage.push_back(*connNode);
          }
      }
    }
    mDist.push_back(stage);
  }
  mDist.erase(mDist.end()-1);
  vector<vector<int> >::iterator itStages; // print profile
  cout << "mDist[";
  for (itStages=mDist.begin(); mDist.end()!=itStages; itStages++)
    cout << itStages->size() << ",";
  cout << "]\n";
}

void    RubberbandOrder::Calculate()
{
  assert(mNumNodes==mCurNum);
  assert(mNumNodes==(int)mNameToNumber.size());
  MakeDistances(1);
  assert(mConnected.end()==find(mConnected.begin()+1,mConnected.end(),0)); // we don`t handle unconnected graphs, yet.
  MakeDistances(mDist.back()[0]);
  // Now we have a top node and calculate a bottom node.
  mTopNode=mDist.back()[0];
  MakeDistances(mTopNode);
  mBotNode=mDist.back()[0];
  MakeOrder();
}

void    RubberbandOrder::MakeOrder()
  // this does the actual calculation of an order.
  // This is a very stupid algoritm, but we will change this only if
  // the order is successful.
{
  mPos=vector<double>(mNumNodes+1); mPos[mTopNode]=1; mPos[mBotNode]=-1;
  // we calculate iteratively the stabilized position of a net of rubber bands of
  // strength mWeight spanned between connected nodes with fixed top and bot nodes.
  double change=1000;
  cout << "Change: "; cout.flush();
  for (int kn=0; (kn<10)||(change>0.1/mNumNodes); kn++) {
    change=0;
    for (int i=0; i<mNumNodes*3; i++) {
      int node=random()%mNumNodes+1; // the node we are going to move
      if (node==mTopNode || node==mBotNode) continue;
      double newpos=MoveNode(node); // calc. the new position of the node.
      change+=fabs(newpos-mPos[node]);
      mPos[node]=newpos;
    }
  cout << change/mNumNodes << ", "; cout.flush();
  }
  cout << "\n"; cout.flush();
  // now we sort the elements according to mPos.
  vector<int>   srt_ind(mNumNodes);
  iota(srt_ind.begin(),srt_ind.end(),0);
  mPos["asdr"];
  sort(srt_ind.begin(),srt_ind.end(),
       compose2(less<double>(),
                bindthis(&vector<double>::operator[],&mPos),
                bindthis(&vector<double>::operator[],&mPos)));
  cout << "sorted: ";
  transform(srt_ind.begin(),srt_ind.end(),ostream_iterator<double>(cout,", "),
            bindthis(&vector<double>::operator[],&mPos));
}

double  RubberbandOrder::MoveNode(int node) 
{ // calculate new position of node
  int   sumweights=0;
  double        res=0;
  vector<int>::iterator itConns; // iterate over connected Nodes
  for (itConns=mEdge[node].begin(); mEdge[node].end()!=itConns; itConns++ ) {
    int weight=mWeight[make_pair(node,*itConns)];
    sumweights+=weight;
    res+=weight*mPos[*itConns];
  }
  return res/sumweights;
}


string  makestring(long n)
{
  char  res[50];
  sprintf(res,"N%ld",n);
  return string(res);
}

int     main()
{
  const int nodes=1000,edges=5000;
  srandom(time(NULL));
  RubberbandOrder        test;
  for (int i=0; i<edges; i++) {
    string node1,node2;
    node1=makestring(random()%nodes);
    do { node2=makestring(random()%nodes); } while(node1==node2);
    test.InsertEdge(make_pair(node1,node2));
  }
  printit(test.Result());
  return 0;
}

