// Hey Emacs! This is -*- c++ -*- ! :-)
// Term.cc	-- Hans-Peter Stoerr Thu Jan 14 1999

#include <strstream>
#include <assert.h>
#include <algo.h>

#include "general.h"

#include "Term.h"
#include "Term_func.h"          // includes Term.h as well
#include "misc/mystl.h"

#pragma implementation

//================================================================
// Term Implementation

Term::Term(TermRep* p)
{
  rep=p;
  p->refs++;
}

bool operator<(const Term& t1, const Term& t2) 
{
  return t1.Cmp(t2)<0;
}

int Term::Cmp(const Term& t2) const
{
  if(Type()<t2.Type()) return -1;
  if(Type()>t2.Type()) return 1;
  return rep->Cmp(*t2.rep);
}

void Term::Printit(ostream& s) const
{
  if(rep) rep->Printit(s);
  else s << "<NULL>";
}

string Term::Describe() const
{
  if (NULL==rep) return "<NULL>";
  if (Atom==Type()) return ((TermAtom*)rep)->Describe();
  strstream s;
  s << *this;
  return string(s.str());
}

// ================================================================
// Subclasses of TermRep for representation of various kinds of terms

TermAc1Binary::TermAc1Binary(Term::TermType thetype) : TermRep(thetype)
{
  switch(type) {
  case Term::And:       name="*"; break;
  case Term::Or:        name="+"; break;
  default:              IMPOSSIBLE;
  }
}

void TermAc1Binary::Printit(ostream& s) const
{
  s << name << args;
}

Term MakeTerm(Term::TermType type, TermSet args) 
{
  TermAc1Binary* rep=new TermAc1Binary(type);
  rep->args=args;
  return rep;
}

int   TermAc1Binary::Cmp(const TermRep& t) const
{
  const TermAc1Binary &rt=*(const TermAc1Binary *)&t;
  return lexicographical_compare_3way(args.begin(), args.end(),
                                      rt.args.begin(), rt.args.end());
}

// _______________________________________________________________

TermUnary::TermUnary(Term::TermType thetype) : TermRep(thetype)
{
  switch(type) {
  case Term::Not:       name="-"; break;
  case Term::Special:    name="Special"; break;
  default:              IMPOSSIBLE;
  }
}

void TermUnary::Printit(ostream& s) const
{
  s << name << " " << arg;
}

Term MakeTerm(Term::TermType type, Term arg)
{
  switch(type) {
  case Term::Not:
    {
      TermUnary* rep=new TermUnary(type);
      rep->arg=arg;
      return rep;
    }
  default:
    IMPOSSIBLE; return (TermUnary*)NULL;
  }
}

Term MakeTerm(Term::TermType type, string fname, Term arg)
{
  switch(type) {
  case Term::Special: 
    {
      TermUnary* rep=new TermUnary(type);
      rep->arg=arg;
      rep->name=fname;
      return rep;
    }
  default:
    IMPOSSIBLE; return (TermUnary*)NULL;
  }
}

int   TermUnary::Cmp(const TermRep& t) const
{
  const TermUnary &rt=*(const TermUnary*)&t;
  if (name<rt.name) return -1;
  if (name>rt.name) return 1;
  return arg.Cmp(rt.arg);
}

// _______________________________________________________________

TermBinary::TermBinary(Term::TermType thetype) : TermRep(thetype)
{
  switch(type) {
  case Term::Imply:     name="->"; break;
  case Term::When:      name="when"; break;
  default:              IMPOSSIBLE;
  }
}

void TermBinary::Printit(ostream& s) const
{
  s << name << "(" << arg1 << ", "
    << arg2 << ")";
}

int   TermBinary::Cmp(const TermRep& t) const
{
  const TermBinary &rt=*(const TermBinary *)&t;
  int res=arg1.Cmp(rt.arg1);
  if(0!=res) return res;
  return arg2.Cmp(rt.arg2);
}

Term MakeTerm(Term::TermType type, Term arg1, Term arg2)
{
  switch (type) {
  case(Term::Imply):
  case(Term::When): 
    {
      TermBinary* rep=new TermBinary(type);
      rep->arg1=arg1; rep->arg2=arg2;
      return rep;
    }
  case(Term::And):
  case(Term::Or):
    {
      TermSet     args; args.insert(arg1); args.insert(arg2);
      return MakeTerm(type,args);
    }
  default: IMPOSSIBLE;          // not implemented
  }
  return 0;                     // pacifier for the compiler; is never reached.
}

// _______________________________________________________________

TermAtom::TermAtom(Term::TermType thetype) : TermRep(thetype)
{
  assert(Term::Atom==type);
}

void TermAtom::Printit(ostream& s) const
{
  s << name << "(" << args << ")";
}

int   TermAtom::Cmp(const TermRep& t) const
{
  const TermAtom &rt=*(const TermAtom *)&t;
  int res=name.compare(rt.name);
  if(0!=res) return res;
  return lexicographical_compare_3way(args.begin(), args.end(),
                                      rt.args.begin(), rt.args.end());
}

string TermAtom::Describe() const
{
  string aname=name+"(";
  StringList::const_iterator it=args.begin();
  if (args.end()!=it) aname=aname+*(it++);
  while (args.end()!=it) aname=aname+","+*(it++);
  aname=aname+")";
  return aname;
}

Term MakeTerm(string predicate, StringList args)
{
  TermAtom* rep=new TermAtom(Term::Atom);
  rep->name=predicate; rep->args=args;
  return rep;
}

// _______________________________________________________________

TermQuantor::TermQuantor(Term::TermType thetype) : TermRep(thetype)
{
  switch(type) {
  case Term::Forall:    name="A"; break;
  case Term::Exists:    name="E"; break;
  default:              IMPOSSIBLE;
  }
}

void TermQuantor::Printit(ostream& s) const
{
  StringStringMap::const_iterator it=vars.begin();
  s << name << "(";
  if(it!=vars.end()){
    s << (*it).first << " - " << (*it).second;
    it++;
  }
  while(it!=vars.end()) {
    s << " " << (*it).first << " - " << (*it).second;
    it++;
  }
  s << ") " << arg;  
}

static int cmp(const pair<string,string>& p1, const pair<string,string>& p2)
{
  int res=p1.first.compare(p2.first);
  if(0!=res) return res;
  return p1.second.compare(p2.second);
} 

static int lexcmp(const StringStringMap& m1, const StringStringMap& m2)
{
  StringStringMap::const_iterator it1=m1.begin();
  StringStringMap::const_iterator it2=m2.begin();
  int res=0;
  while( (m1.end()!=it1) && (m2.end()!=it2) ) {
    res=cmp(*it1,*it2);
    if (0!=res) return res;
    ++it1; ++it2;
  }
  if ( m1.end() == it1 )
    return !( m2.end() == it2 );
  return -1;
}

int   TermQuantor::Cmp(const TermRep& t) const
{
  const TermQuantor &rt=*(const TermQuantor *)&t;
  int res=arg.Cmp(rt.arg);
  if(0!=res) return res;
  return lexcmp(vars,rt.vars);
//   return lexicographical_compare_3way(vars.begin(), vars.end(),
//                                       rt.vars.begin(), rt.vars.end());
}

Term MakeTerm(Term::TermType type, StringStringMap vars, Term arg)
{
  TermQuantor* rep=new TermQuantor(type);
  rep->arg=arg;
  rep->vars=vars;
  return rep;
}

// _______________________________________________________________

TermEqual::TermEqual(Term::TermType thetype) : TermRep(thetype)
{
  name="=";
  assert(Term::Equal==type);
}

void TermEqual::Printit(ostream& s) const
{
  s << name << "(" << arg1 << ", " << arg2 << ")";
}

int   TermEqual::Cmp(const TermRep& t) const
{
  const TermEqual &rt=*(const TermEqual *)&t;
  int res=arg1.compare(rt.arg1);
  if(0!=res) return res;
  return arg2.compare(rt.arg2);
}


Term MakeTerm(Term::TermType type, string arg1, string arg2)
{
  TermEqual* rep=new TermEqual(type);
  rep->arg1=arg1; rep->arg2=arg2;
  return rep;
}

// ================================================================
void RecursiveVoidTermFunction::NotFunc(const TermUnary& t)
{
  return Call(t.arg);
}

void RecursiveVoidTermFunction::AndFunc(const TermAc1Binary& t)
{
  for_each(t.args.begin(), t.args.end(), funcref1(*this));
}
           
void RecursiveVoidTermFunction::OrFunc(const TermAc1Binary& t)
{
  for_each(t.args.begin(), t.args.end(), funcref1(*this));
}

void RecursiveVoidTermFunction::ImplyFunc(const TermBinary& t)
{
  Call(t.arg1); Call(t.arg2);
}

void RecursiveVoidTermFunction::WhenFunc(const TermBinary& t)
{
  Call(t.arg1); Call(t.arg2);
}

void RecursiveVoidTermFunction::SpecialFunc(const TermUnary& t)
{
  return Call(t.arg);
}

void RecursiveVoidTermFunction::ForallFunc(const TermQuantor& t)
{
  Call(t.arg);
}

void RecursiveVoidTermFunction::ExistsFunc(const TermQuantor& t)
{
  Call(t.arg);
}

void RecursiveVoidTermFunction::EqualFunc(const TermEqual&)
{
  NOOP;
}

void RecursiveVoidTermFunction::AtomFunc(const TermAtom&)
{
  NOOP;
}

void RecursiveVoidTermFunction::NullFunc()
{
  NOOP;
}

