/* Term_func.h -*- c++ -*- Hans-Peter Stoerr Thu Jan 28 1999
 * This file defines some auxilliary classes for the representation of Terms
 * (TermAc1Binary, TermUnary, TermBinary, TermQuantor, TermAtom) which are
 * TermRep, and defines a class TermFunction for the definition of case based
 * functions operating on Terms. (TermFunction automatically selects the right
 * function to call depending on the term type.
 */

#ifndef _TERM_FUNC_H
#define _TERM_FUNC_H

#include <functional>

#include "Term.h"

// ================================================================
// TermRep Subclasses

class TermAc1Binary : public TermRep
// associative and commutative binary operator: And, Or
{
public:
  TermSet       args;
  
  TermAc1Binary(Term::TermType thetype);
  virtual ~TermAc1Binary() {};

  int   Cmp(const TermRep& t) const;
protected:
  virtual void Printit(ostream& s) const;
};

class TermUnary : public TermRep
// Negation
{
public:
  Term          arg;

  TermUnary(Term::TermType thetype);
  virtual ~TermUnary() {};

  int   Cmp(const TermRep& t) const;

protected:
  virtual void Printit(ostream& s) const;
};

class TermBinary : public TermRep
// Implication
{
public:
  Term          arg1,arg2;
  
  TermBinary(Term::TermType thetype);
  virtual ~TermBinary() {};

  int   Cmp(const TermRep& t) const;

protected:
  virtual void Printit(ostream& s) const;
};

class TermQuantor : public TermRep
{
public:
  StringStringMap vars;
  Term          arg;
  
  TermQuantor(Term::TermType thetype);
  virtual ~TermQuantor() {};

  int   Cmp(const TermRep& t) const;

protected:
  virtual void Printit(ostream& s) const;
};

class TermAtom : public TermRep
// an Atom :-)
{
public:
  StringList    args;
  
  TermAtom(Term::TermType thetype);
  virtual ~TermAtom() {};

  int   Cmp(const TermRep& t) const;
  string Describe() const;

protected:
  virtual void Printit(ostream& s) const;
};

class TermEqual : public TermRep
// an Equality between two args
{
public:
  string        arg1,arg2;
  
  TermEqual(Term::TermType thetype);
  virtual ~TermEqual() {};

  int   Cmp(const TermRep& t) const;

protected:
  virtual void Printit(ostream& s) const;
};

// ================================================================
// TermFunction Class

template <class Resulttype>
class TermFunction : public unary_function<Term, Resulttype>,
                     public Object
{
public:
  virtual       ~TermFunction(){};

  Resulttype    operator()(const Term&);
  Resulttype    Call(const Term& t) { return (*this)(t); } // just an alias for ()

protected:
  TermFunction(){};
  virtual Resulttype    NotFunc(const TermUnary&)=0;
  virtual Resulttype    AndFunc(const TermAc1Binary&)=0;
  virtual Resulttype    OrFunc(const TermAc1Binary&)=0;
  virtual Resulttype    ImplyFunc(const TermBinary&)=0;
  virtual Resulttype    WhenFunc(const TermBinary&)=0;
  virtual Resulttype    SpecialFunc(const TermUnary&)=0;
  virtual Resulttype    ForallFunc(const TermQuantor&)=0;
  virtual Resulttype    ExistsFunc(const TermQuantor&)=0;
  virtual Resulttype    AtomFunc(const TermAtom&)=0;
  virtual Resulttype    EqualFunc(const TermEqual&)=0;
  virtual Resulttype    NullFunc()=0;
};

template <class Argtype, class Resulttype>
class TermBinaryFunction : public binary_function<Term, Argtype, Resulttype>,
                           public Object
{
public:
  virtual       ~TermBinaryFunction(){};

  Resulttype    operator()(const Term&, Argtype);
  // just an alias for () :
  Resulttype    Call(const Term& t, Argtype arg){ return (*this)(t,arg); }

protected:
  TermBinaryFunction(){};
  virtual Resulttype    NotFunc(const TermUnary&, Argtype)=0;
  virtual Resulttype    AndFunc(const TermAc1Binary&, Argtype)=0;
  virtual Resulttype    OrFunc(const TermAc1Binary&, Argtype)=0;
  virtual Resulttype    ImplyFunc(const TermBinary&, Argtype)=0;
  virtual Resulttype    WhenFunc(const TermBinary&, Argtype)=0;
  virtual Resulttype    SpecialFunc(const TermUnary&, Argtype)=0;
  virtual Resulttype    ForallFunc(const TermQuantor&, Argtype)=0;
  virtual Resulttype    ExistsFunc(const TermQuantor&, Argtype)=0;
  virtual Resulttype    AtomFunc(const TermAtom&, Argtype)=0;
  virtual Resulttype    EqualFunc(const TermEqual&, Argtype)=0;
  virtual Resulttype    NullFunc(Argtype)=0;
};

// Termfunction Subclass which just recurses over whole term.

class RecursiveVoidTermFunction : public TermFunction<void> 
{
protected:
  virtual void  NotFunc(const TermUnary&);
  virtual void  AndFunc(const TermAc1Binary&);
  virtual void  OrFunc(const TermAc1Binary&);
  virtual void  ImplyFunc(const TermBinary&);
  virtual void  WhenFunc(const TermBinary&);
  virtual void  SpecialFunc(const TermUnary&);
  virtual void  ForallFunc(const TermQuantor&);
  virtual void  ExistsFunc(const TermQuantor&);
  virtual void  AtomFunc(const TermAtom&);
  virtual void  EqualFunc(const TermEqual&);
  virtual void  NullFunc();
};

// ================================================================
// now some implementations which may have to be instantiated.

template <class Resulttype>
Resulttype TermFunction<Resulttype>::operator()(const Term& t)
{
  switch(t.Type()) {
  case Term::Not:       return NotFunc( (const TermUnary&)*t );
  case Term::And:       return AndFunc( (const TermAc1Binary&)*t );
  case Term::Or:        return OrFunc( (const TermAc1Binary&)*t );
  case Term::Imply:     return ImplyFunc( (const TermBinary&)*t );
  case Term::When:      return WhenFunc( (const TermBinary&)*t );
  case Term::Special:   return SpecialFunc( (const TermUnary&)*t );
  case Term::Forall:    return ForallFunc( (const TermQuantor&)*t );
  case Term::Exists:    return ExistsFunc( (const TermQuantor&)*t );
  case Term::Atom:      return AtomFunc( (const TermAtom&)*t );
  case Term::Equal:     return EqualFunc( (const TermEqual&)*t );
  case Term::Null:      return NullFunc();
  default:              IMPOSSIBLE; return Resulttype();
  }
}

template <class Argtype, class Resulttype>
Resulttype TermBinaryFunction<Argtype,Resulttype>::operator()(const Term& t, Argtype arg)
{
  switch(t.Type()) {
  case Term::Not:       return NotFunc( (const TermUnary&)*t, arg ); break;
  case Term::And:       return AndFunc( (const TermAc1Binary&)*t, arg ); break;
  case Term::Or:        return OrFunc( (const TermAc1Binary&)*t, arg ); break;
  case Term::Imply:     return ImplyFunc( (const TermBinary&)*t, arg ); break;
  case Term::When:      return WhenFunc( (const TermBinary&)*t, arg ); break;
  case Term::Special:   return SpecialFunc( (const TermUnary&)*t, arg ); break;
  case Term::Forall:    return ForallFunc( (const TermQuantor&)*t, arg ); break;
  case Term::Exists:    return ExistsFunc( (const TermQuantor&)*t, arg ); break;
  case Term::Atom:      return AtomFunc( (const TermAtom&)*t, arg ); break;
  case Term::Equal:     return EqualFunc( (const TermEqual&)*t, arg ); break;
  case Term::Null:      return NullFunc( arg ); break;
  default:              IMPOSSIBLE; break;
  }
  return Resulttype();
}

#endif
