/* mystl.h	-- Hans-Peter Stoerr Fri Jan 29 1999
 * $Id: mystl.h,v 2.2 2003/12/04 07:23:37 stoerr Exp $
 * some STL-like stuff (algorithms, functionals, etc.)
 */

#ifndef _MYSTL_H
#define _MYSTL_H

#include <memory>
using namespace std;

// ================================================================
// Accessorfunctions to pair:

template <class T1, class T2>
T1& first(pair<T1,T2>& p) { return p.first; }
  
template <class T1, class T2>
T1& second(pair<T1,T2>& p) { return p.second; }  

// ================================================================
// call f(*(first1+i), *(first2+i))
// for i in [0, min(last1-first1, last2-first2) )

template <class InputIterator1, class InputIterator2, class Function>
Function for_each2(InputIterator1 first1, InputIterator1 last1,
                     InputIterator2 first2, InputIterator2 last2,
                     Function f) {
  for ( ; (first1 != last1) && (first2 != last2); ++first1, ++first2)
    f(*first1,*first2);
  return f;
}

// ================================================================
// copy all elements satisfying Predicate (untested.)

template <class InputIterator, class OutputIterator, class Predicate>
inline OutputIterator copy_if(InputIterator first, InputIterator last,
                           OutputIterator result, Predicate pred)
{
  for ( ; first != last; ++first)
    if ( pred(*first) ) {
      *result = *first; ++result;
    }
  return result;
}

// ================================================================
// passes all elements satisfying pred through op and to result...

template <class InputIterator, class OutputIterator,
  class UnaryOperation, class UnaryPredicate>
OutputIterator transform_if(InputIterator first, InputIterator last,
                            OutputIterator result, UnaryOperation op,
                            UnaryPredicate test)
{
  while (first!=last) {
    if (test(*first)) {
      *result = op(*first);
      result++;
    }
    ++ first;
  }
  return result;
}

// ================================================================
// bind implicit this argument of a member function with one arg.
// bindthis1(f,this)(x) is the same as this->f(x).
// untested yet.

template <class Class, class Arg, class Result> 
class binderthis1
  : public unary_function<Arg, Result> {
protected:
    Class *thisvalue;
    Result (Class::* fun)(Arg);
public:
    binderthis1(Result (Class::* thefun)(Arg), Class *thethisvalue)
      : thisvalue(thethisvalue), fun(thefun) {};
    Result operator()(Arg arg) const {
      return (thisvalue->*fun)(arg);
    }
};

template <class Class, class Arg, class Result>
inline binderthis1<Class,Arg,Result>
bindthis(Result (Class::* fun)(Arg), Class* thisvalue)
{
  return binderthis1<Class,Arg,Result>(fun, thisvalue);
}

//________________________________________________________________
// const version

template <class Class, class Arg, class Result> 
class binderthis1const
  : public unary_function<Arg, Result> {
protected:
    const Class *thisvalue;
    Result (Class::* fun)(Arg) const;
public:
    binderthis1const(Result (Class::* thefun)(Arg) const, const Class *thethisvalue)
      : thisvalue(thethisvalue), fun(thefun) {};
    Result operator()(Arg arg) const {
      return (thisvalue->*fun)(arg);
    }
};

template <class Class, class Arg, class Result>
inline binderthis1const<Class,Arg,Result>
bindthis(Result (Class::* fun)(Arg) const, const Class* thisvalue)
{
  return binderthis1const<Class,Arg,Result>(fun, thisvalue);
}

// ================================================================
// bind implicit this argument of a member function with two args.
// bindthis1(f,this)(x,y) is the same as this->f(x,y).

template <class Class, class Arg1, class Arg2, class Result> 
class binderthis2
  : public binary_function<Arg1, Arg2, Result> {
protected:
    Result (Class::* fun)(Arg1,Arg2);
    Class *thisvalue;
public:
    binderthis2(Result (Class::* thefun)(Arg1,Arg2), Class *thethisvalue)
      : fun(thefun), thisvalue(thethisvalue) {};
    Result operator()(Arg1 arg1,Arg2 arg2) const {
      return (thisvalue->*fun)(arg1, arg2);
    }
};

template <class Class, class Arg1, class Arg2, class Result>
inline binderthis2<Class,Arg1,Arg2,Result>
bindthis(Result (Class::* fun)(Arg1,Arg2), Class* thisvalue)
{
  return binderthis2<Class,Arg1,Arg2,Result>(fun, thisvalue);
}

// ________________________________________________________________
// const version

template <class Class, class Arg1, class Arg2, class Result> 
class binderthis2const
  : public binary_function<Arg1, Arg2, Result> {
protected:
    Result (Class::* fun)(Arg1,Arg2) const;
    const Class *thisvalue;
public:
    binderthis2const(Result (Class::* thefun)(Arg1,Arg2) const,
                     const Class *thethisvalue)
      : fun(thefun), thisvalue(thethisvalue) {};
    Result operator()(Arg1 arg1,Arg2 arg2) const {
      return (thisvalue->*fun)(arg1, arg2);
    }
};

template <class Class, class Arg1, class Arg2, class Result>
inline binderthis2const<Class,Arg1,Arg2,Result>
bindthis(Result (Class::* fun)(Arg1,Arg2) const, Class* thisvalue)
{
  return binderthis2const<Class,Arg1,Arg2,Result>(fun, thisvalue);
}

// ================================================================
// wrapper for function object so we can pass objects with virtual functions
// to algorithms.

template <class Operation>
class unary_functionref
: public unary_function<typename Operation::argument_type,
  typename Operation::result_type> 
{
public:
  typedef typename Operation::argument_type Arg;
  typedef typename Operation::result_type Result;
  unary_functionref(Operation& theref) : ref(theref) {}
  Result operator() (const Arg& arg) { return ref(arg); }
private:
  Operation& ref;
};

template <class Operation>
inline unary_functionref<Operation> funcref1(Operation& obj)
{
  return unary_functionref<Operation>(obj);
}

// ================================================================
// this works like accumulate but inserts a separator between elements
// i.e. the result is *first+sep+*(first+1)+sep+...+sep+*(last-1)

template <class InputIterator, class T>
T accumulatesep(InputIterator first, const InputIterator last, T init, const T sep) {
  if( last==first ) return init;
  init = init + *(first++);
  for ( ; first != last; ++first)
    init = init + sep + *first;
  return init;
}

// ================================================================
// generator that returns some value on the first call, then another.

template<class T>
class FirsttimeNexttime
{
public:
  FirsttimeNexttime(const T theFirsttime, const T theNexttime)
    : mIsFirsttime(TRUE), mFirsttimeVal(theFirsttime), mNexttimeVal(theNexttime) { }
  T operator() () {
    bool   is_firsttime=mIsFirsttime; mIsFirsttime=FALSE;
    return is_firsttime ? mFirsttimeVal : mNexttimeVal;
  }
private:
  bool  mIsFirsttime;
  const T       mFirsttimeVal;
  const T       mNexttimeVal;
};

#endif

