/* pddldomain.h	-*- c++ -*- Hans-Peter Stoerr Thu Jan  7 1999 */

#ifndef _PDDLDOMAIN_H
#define _PDDLDOMAIN_H

#include <string>
#include <list>
#include <set>
#include <map>

#include "Printable.h"
#include "Term.h"

typedef set<string, less<string> > StringSet;
typedef list<string> StringList;
typedef map<string, string, less<string> > StringStringMap;
typedef map<string, StringList, less<string> > StringStringListMap;
typedef set<Term, less<Term> > TermSet;
typedef pair<string, string> StringStringPair;
typedef list<StringStringPair> StringStringPairList;

class PddlDatabase;
class PddlDomain;
class PddlSituation;

typedef vector<string>  Atomtype;
typedef map<string, Atomtype, less<string> > StringAtomtypeMap;

// ================================================================
// superclass of all Pddl items remembering definition position.

class PddlItem : public Printable
{
public:
  PddlItem();
  virtual ~PddlItem();
  
  string        filename;
  unsigned      linenumber;
};

// ================================================================

class PddlAction : public PddlItem
{
  friend int yyparse(void);
  friend class PddlDomain;
public:
  string                name;
  StringStringPairList  parameters;
  StringStringMap       vars;
  Term                  precondition;
  Term                  effect;

  PddlAction();
  virtual ~PddlAction();

protected:
  // for use by yyparse; these operations delete the pointer afterwards.
  PddlAction(string *thename);
  void  SetParameters(StringStringPairList *parms);
  void  SetPrecondition(Term* pre);
  void  SetEffect(Term* eff);
  void  SetPosEffect(TermSet* eff);
  void  SetNegEffect(TermSet* eff);
  void  SetVars(StringStringMap *v);

protected:
  virtual void Printit(ostream& s) const;
};

// bool  operator<(const PddlAction& a, const PddlAction& b); // for use with set<...>

typedef map<string,PddlAction,less<string> > StringPddlActionMap;

// ================================================================
class PddlAxiom : public PddlItem
{
  friend int yyparse(void);
public:
  StringStringMap vars;
  Term          context;
  Term          implicand;

protected:
  // for use by yyparse; these operations delete the pointer afterwards.
  void  SetVars(StringStringMap *s);
  void  SetContext(Term *c);
  void  SetImplicand(Term *i);

  virtual void Printit(ostream& s) const;
};

bool  operator<(const PddlAxiom& a, const PddlAxiom& b); // for use with set<...>
  
typedef set<PddlAxiom, less<PddlAxiom> > PddlAxiomSet;

// ================================================================

class PddlProblem: public PddlItem
{
  friend int yyparse(void);
  friend class PddlDatabase;
public:
  const string  name;
  string        domainname;
  StringSet     requirements;
  StringStringMap constants;
  TermSet       init;
  Term          goal;
  int           serial, parallel;

protected:
  virtual ~PddlProblem();
  PddlProblem(const PddlProblem&);// unimplemented, can't be done.
  PddlProblem& operator=(const PddlProblem&); // unimplemented, can't be done.

  // for use by yyparse; these operations delete the pointer afterwards.
  PddlProblem(string *name);
  void  Requires(StringSet*);
  void  BasedOnSituation(const PddlSituation& sit);
  void  SetObjects(StringStringMap*);
  void  SetInit(TermSet*);
  void  SetGoal(Term*);

  virtual void Printit(ostream& s) const;
};

// ================================================================
// This is not really implemented yet. Just a place holder.
class PddlSituation: public PddlItem
{
  friend int yyparse(void);
  friend class PddlDatabase;
public:
  const string  name;
  string        domainname;
  StringStringMap constants;
  TermSet       init;

protected:
  // for use by yyparse; these operations delete the pointer afterwards.
  PddlSituation(string* thename);
  PddlSituation(const PddlSituation&); // unimplemented, can't be done.
  virtual ~PddlSituation();
  PddlSituation& operator=(const PddlSituation*); // unimplemented, can't be done.

  void  SetDomainname(string *);
  void  SetConstants(StringStringMap *);
  void  SetInit(TermSet *);
  
  virtual void Printit(ostream& s) const;  
};

// ================================================================

class PddlDomain: public PddlItem
{
  friend int yyparse(void);
  friend class PddlDatabase;
public:
  const string  name;
  StringSet     requirements;
  StringStringMap directsupertypes;
  StringStringMap constants;
  StringAtomtypeMap atomtypes;
  TermSet       safeties;
  TermSet       constraints;
  TermSet       timeless;
  PddlAxiomSet      axioms;
  StringPddlActionMap actions;

protected:
  virtual ~PddlDomain();
  // for use by yyparse; these operations delete the pointer afterwards.
  PddlDomain(string *thename);
  void  Extends(string *);
  void  Requires(StringSet *);
  void  SetSubtypes(StringStringMap *);
  void  SetConstants(StringStringMap *);
  void  SetAtomtypes(StringAtomtypeMap *);
  void  AddSafety(Term *s);
  void  AddConstraint(Term *s);
  void  SetTimeless(TermSet *);
  void  AddAxiom(PddlAxiom *ax);
  void  AddAction(PddlAction *ac);

  virtual void Printit(ostream& s) const;
};

// ================================================================
// class for storing domains, problems and situations

class   PddlDatabase : public Printable
{
  friend int yyparse(void);
public:
  PddlDatabase();
  virtual ~PddlDatabase();

  // destruction of Database object kills domains etc. as well.
  const PddlDomain* Domain(const string) const; // returns first domain on ""
  const PddlProblem* Problem(const string) const;  // returns first problem on ""
  // returns first problem of domain on problem=="":
  const PddlProblem* Problem(const string domain,const string problem) const;
  const PddlSituation* Situation(const string) const;

  bool  ParseFile(const char*); // for the moment: aborts if error. returns true on success.
  bool  ParseFiles(const vector<string>); // for the moment: aborts if error. returns true on success.
  void  ListProblems() const;   // print all found problems to stdout.
  void  ListDomains() const;    // print all found domains to stdout.
  void  ListProblemsWithDomains() const; // print all found problems with their domains to stdout.
  
protected:
  void  RegisterDomain(PddlDomain*); // for use by yyparse
  void  RegisterProblem(PddlProblem*); // for use by yyparse
  void  RegisterSituation(PddlSituation*); // for use by yyparse

  map<string, PddlDomain*, less<string> > domains;
  map<string, PddlProblem*, less<string> > problems;
  map<string, PddlSituation*, less<string> > situations;

  virtual void Printit(ostream& s) const;
private:
  PddlDatabase(const PddlDatabase&); // unimplemented, can't be done.
  PddlDatabase& operator=(const PddlDatabase&); // unimplemented, can't be done.
};

#endif
